@@ -1,5 +1,5 @@
 #
-# (C) Copyright 2003-2004 Jacek Konieczny <jajcus@jajcus.net>
+# (C) Copyright 2003-2010 Jacek Konieczny <jajcus@jajcus.net>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License Version
@@ -21,7 +21,6 @@ Normative reference:
   - `RFC 3920 <http://www.ietf.org/rfc/rfc3920.txt>`__
 """
 
-__revision__="$Id: stanzaprocessor.py 601 2005-08-31 17:31:42Z jajcus $"
 __docformat__="restructuredtext en"
 
 import libxml2
@@ -29,6 +28,8 @@ import logging
 import threading
 
 from pyxmpp.expdict import ExpiringDictionary
+from pyxmpp.exceptions import ProtocolError, BadRequestProtocolError, FeatureNotImplementedProtocolError
+from pyxmpp.stanza import Stanza
 
 class StanzaProcessor:
     """Universal stanza handler/router class.
@@ -59,7 +60,34 @@ class StanzaProcessor:
         self.__logger=logging.getLogger("pyxmpp.Stream")
         self.lock=threading.RLock()
 
-    def process_iq(self,stanza):
+    def process_response(self, response):
+        """Examines out the response returned by a stanza handler and sends all
+        stanzas provided.
+
+        :Returns:
+           - `True`: if `response` is `Stanza`, iterable or `True` (meaning the stanza was processed).
+           - `False`: when `response` is `False` or `None`
+        :returntype: `bool`
+        """
+
+        if response is None or response is False:
+            return False
+
+        if isinstance(response, Stanza):
+            self.send(response)
+            return True
+
+        try:
+            response = iter(response)
+        except TypeError:
+            return bool(response)
+
+        for stanza in response:
+            if isinstance(stanza, Stanza):
+                self.send(stanza)
+        return True
+
+    def process_iq(self, stanza):
         """Process IQ stanza received.
 
         :Parameters:
@@ -78,39 +106,46 @@ class StanzaProcessor:
                 ufr=fr.as_unicode()
             else:
                 ufr=None
-            if self._iq_response_handlers.has_key((sid,ufr)):
-                key=(sid,ufr)
-            elif ( (fr==self.peer or fr==self.me)
-                    and self._iq_response_handlers.has_key((sid,None))):
-                key=(sid,None)
+            res_handler = err_handler = None
+            try:
+                res_handler, err_handler = self._iq_response_handlers.pop((sid,ufr))
+            except KeyError:
+                if ( (fr==self.peer or fr==self.me or fr==self.me.bare()) ):
+                    try:
+                        res_handler, err_handler = self._iq_response_handlers.pop((sid,None))
+                    except KeyError:
+                        pass
+                if None is res_handler is err_handler:
+                    return False
+            if typ=="result":
+                response = res_handler(stanza)
             else:
-                return 0
-            res_handler,err_handler=self._iq_response_handlers[key]
-            if stanza.get_type()=="result":
-                res_handler(stanza)
-            else:
-                err_handler(stanza)
-            del self._iq_response_handlers[key]
-            return 1
+                response = err_handler(stanza)
+            self.process_response(response)
+            return True
 
         q=stanza.get_query()
         if not q:
-            r=stanza.make_error_response("bad-request")
-            self.send(r)
-            return 1
+            raise BadRequestProtocolError, "Stanza with no child element"
         el=q.name
         ns=q.ns().getContent()
 
-        if typ=="get" and self._iq_get_handlers.has_key((el,ns)):
-            self._iq_get_handlers[(el,ns)](stanza)
-            return 1
-        elif typ=="set" and self._iq_set_handlers.has_key((el,ns)):
-            self._iq_set_handlers[(el,ns)](stanza)
-            return 1
+        if typ=="get":
+            if self._iq_get_handlers.has_key((el,ns)):
+                response = self._iq_get_handlers[(el,ns)](stanza)
+                self.process_response(response)
+                return True
+            else:
+                raise FeatureNotImplementedProtocolError, "Not implemented"
+        elif typ=="set":
+            if self._iq_set_handlers.has_key((el,ns)):
+                response = self._iq_set_handlers[(el,ns)](stanza)
+                self.process_response(response)
+                return True
+            else:
+                raise FeatureNotImplementedProtocolError, "Not implemented"
         else:
-            r=stanza.make_error_response("feature-not-implemented")
-            self.send(r)
-            return 1
+            raise BadRequestProtocolError, "Unknown IQ stanza type"
 
     def __try_handlers(self,handler_list,typ,stanza):
         """ Search the handler list for handlers matching
@@ -148,7 +183,8 @@ class StanzaProcessor:
                 continue
             if ns is not None and ns not in namespaces:
                 continue
-            if handler(stanza):
+            response = handler(stanza)
+            if self.process_response(response):
                 return True
         return False
 
@@ -165,14 +201,14 @@ class StanzaProcessor:
 
         if not self.initiator and not self.peer_authenticated:
             self.__logger.debug("Ignoring message - peer not authenticated yet")
-            return 1
+            return True
 
         typ=stanza.get_type()
         if self.__try_handlers(self._message_handlers,typ,stanza):
-            return 1
+            return True
         if typ!="error":
             return self.__try_handlers(self._message_handlers,"normal",stanza)
-        return 0
+        return False
 
     def process_presence(self,stanza):
         """Process presence stanza.
@@ -185,7 +221,7 @@ class StanzaProcessor:
 
         if not self.initiator and not self.peer_authenticated:
             self.__logger.debug("Ignoring presence - peer not authenticated yet")
-            return 1
+            return True
 
         typ=stanza.get_type()
         if not typ:
@@ -206,9 +242,9 @@ class StanzaProcessor:
             - `stanza`: presence stanza to be processed
         """
         if stanza.get_type() not in ("error","result"):
-            r=stanza.make_error_response("recipient-unavailable")
+            r = stanza.make_error_response("recipient-unavailable")
             self.send(r)
-        return 1
+        return True
 
     def process_stanza(self,stanza):
         """Process stanza received from the stream.
@@ -231,15 +267,25 @@ class StanzaProcessor:
         if not self.process_all_stanzas and to and to!=self.me and to.bare()!=self.me.bare():
             return self.route_stanza(stanza)
 
-        if stanza.stanza_type=="iq":
-            if self.process_iq(stanza):
-                return True
-        elif stanza.stanza_type=="message":
-            if self.process_message(stanza):
-                return True
-        elif stanza.stanza_type=="presence":
-            if self.process_presence(stanza):
-                return True
+        try:
+            if stanza.stanza_type=="iq":
+                if self.process_iq(stanza):
+                    return True
+            elif stanza.stanza_type=="message":
+                if self.process_message(stanza):
+                    return True
+            elif stanza.stanza_type=="presence":
+                if self.process_presence(stanza):
+                    return True
+        except ProtocolError, e:
+            typ = stanza.get_type()
+            if typ != 'error' and (typ != 'result' or stanza.stanza_type != 'iq'):
+                r = stanza.make_error_response(e.xmpp_name)
+                self.send(r)
+                e.log_reported()
+            else:
+                e.log_ignored()
+
         self.__logger.debug("Unhandled %r stanza: %r" % (stanza.stanza_type,stanza.serialize()))
         return False
 
@@ -263,10 +309,14 @@ class StanzaProcessor:
             - `iq`: an IQ stanza
             - `res_handler`: result handler for the stanza. Will be called
               when matching <iq type="result"/> is received. Its only
-              argument will be the stanza received.
+              argument will be the stanza received. The handler may return
+              a stanza or list of stanzas which should be sent in response.
             - `err_handler`: error handler for the stanza. Will be called
               when matching <iq type="error"/> is received. Its only
-              argument will be the stanza received.
+              argument will be the stanza received. The handler may return
+              a stanza or list of stanzas which should be sent in response
+              but this feature should rather not be used (it is better not to
+              respond to 'error' stanzas).
             - `timeout_handler`: timeout handler for the stanza. Will be called
               when no matching <iq type="result"/> or <iq type="error"/> is
               received in next `timeout` seconds. The handler should accept
@@ -303,7 +353,8 @@ class StanzaProcessor:
             - `namespace`: payload element namespace URI
             - `handler`: function to be called when a stanza
               with defined element is received. Its only argument
-              will be the stanza received.
+              will be the stanza received. The handler may return a stanza or
+              list of stanzas which should be sent in response.
 
         Only one handler may be defined per one namespaced element.
         If a handler for the element was already set it will be lost
@@ -337,7 +388,9 @@ class StanzaProcessor:
             - `namespace`: payload element namespace URI
             - `handler`: function to be called when a stanza
               with defined element is received. Its only argument
-              will be the stanza received.
+              will be the stanza received. The handler may return a stanza or
+              list of stanzas which should be sent in response.
+
 
         Only one handler may be defined per one namespaced element.
         If a handler for the element was already set it will be lost
@@ -375,7 +428,7 @@ class StanzaProcessor:
         handler_list.append((priority,typ,namespace,handler))
         handler_list.sort()
 
-    def set_message_handler(self,typ,handler,namespace=None,priority=100):
+    def set_message_handler(self, typ, handler, namespace=None, priority=100):
         """Set a handler for <message/> stanzas.
 
         :Parameters:
@@ -386,15 +439,20 @@ class StanzaProcessor:
               payload (or even with no payload) will match.
             - `priority`: priority value for the handler. Handlers with lower
               priority value are tried first.
+            - `handler`: function to be called when a message stanza
+              with defined type and payload namespace is received. Its only
+              argument will be the stanza received. The handler may return a
+              stanza or list of stanzas which should be sent in response.
 
         Multiple <message /> handlers with the same type/namespace/priority may
         be set. Order of calling handlers with the same priority is not defined.
-        Handlers will be called in priority order until one of them returns True.
+        Handlers will be called in priority order until one of them returns True or
+        any stanza(s) to send (even empty list will do).
         """
         self.lock.acquire()
         try:
             if not typ:
-                typ=="normal"
+                typ = "normal"
             self.__add_handler(self._message_handlers,typ,namespace,priority,handler)
         finally:
             self.lock.release()
@@ -403,15 +461,20 @@ class StanzaProcessor:
         """Set a handler for <presence/> stanzas.
 
         :Parameters:
-            - `typ`: presence type. `None` will be treated the same as "available".
+            - `typ`: presence type. "available" will be treated the same as `None`.
             - `namespace`: payload namespace. If `None` that presence with any
               payload (or even with no payload) will match.
             - `priority`: priority value for the handler. Handlers with lower
               priority value are tried first.
+            - `handler`: function to be called when a presence stanza
+              with defined type and payload namespace is received. Its only
+              argument will be the stanza received. The handler may return a
+              stanza or list of stanzas which should be sent in response.
 
         Multiple <presence /> handlers with the same type/namespace/priority may
         be set. Order of calling handlers with the same priority is not defined.
-        Handlers will be called in priority order until one of them returns True.
+        Handlers will be called in priority order until one of them returns
+        True or any stanza(s) to send (even empty list will do).
         """
         self.lock.acquire()
         try:
